from toposort import toposort
from memoria import HardMemory, Pickler
import graphviz
from .NodeStyle import NodeStyle
from .EdgeStyle import EdgeStyle
from .ComputationGraphNode import ComputationGraphNode
from slytherin.collections import remove_list_duplicates
from revelio import make_dir
from interaction import ProgressBar
import os.path

class ComputationGraph():
	def __init__(self, echo=0, hard_memory_path='computation_graph', node_style=None, edge_style=None, on_disk=False, safe=True):
		"""
		:type echo:
		:type hard_memory_path: str
		:type node_style: NodeStyle or None
		:type safe: bool
		:param safe: when True the nodes are created as safe nodes
		"""
		echo = max(0, echo)
		self._echo = echo
		self._nodes = {}
		self._parent_names = {}
		self._child_names = {}
		self._hard_memory = None
		self._hard_memory_path = hard_memory_path
		self._safe = safe
		original_node_style = NodeStyle(
			colour='gray80',
			text_colour='deepskyblue2',
			fill_colour='gray95',
			text_size=8,
			font='helvetica',
			shape='egg',
			style='filled'

		)
		if node_style is None:
			node_style = original_node_style
		else:
			node_style = node_style.complement(original_node_style)

		self._node_style = node_style

		original_edge_style = EdgeStyle(
			colour='grey60',
			arrow_size=0.5
		)
		if edge_style is None:
			edge_style = original_edge_style
		else:
			edge_style = edge_style.complement(original_edge_style)

		self._edge_style = edge_style
		self._on_disk = on_disk

	def save(self, dir, echo=2):
		echo = max(0, echo)
		make_dir(path=dir, ignore_if_exists=True)
		old_nodes = self.nodes
		new_nodes = {}
		for name, node in old_nodes.items():
			path = os.path.join(dir, f'node_{name}.pickle')
			node.save(path=path, echo=echo-1)
			new_nodes[name] = path
		graph_path = os.path.join(dir, 'graph.pickle')
		self._nodes = new_nodes
		Pickler.save(path=graph_path, obj=self)
		self._nodes = old_nodes

	@staticmethod
	def load(dir, echo=1):
		"""
		:type dir: str
		:rtype: ComputationGraph
		"""
		graph = Pickler.load(path=os.path.join(dir, 'graph.pickle'), echo=echo)
		paths = graph._nodes
		graph._nodes = {}
		for name, path in paths.items():
			graph._nodes[name] = ComputationGraphNode.load(path=path, graph=graph, echo=echo)
		return graph

	@property
	def nodes(self):
		"""
		:rtype: dict[str,ComputationGraphNode]
		"""
		return self._nodes

	def freeze(self, node):
		"""
		:type node: ComputationGraphNode or str
		"""
		if isinstance(node, str):
			node_name = node
		else:
			node_name = node.name
		self.nodes[node_name].freeze()

	def unfreeze(self, node):
		"""
		:type node: ComputationGraphNode or str
		"""
		if isinstance(node, str):
			node_name = node
		else:
			node_name = node.name
		self.nodes[node_name].unfreeze()


	@property
	def node_style(self):
		"""
		:rtype: NodeStyle 
		"""
		return self._node_style

	def get_children(self, name):
		return [self._nodes[child_name] for child_name in self._child_names[name]]

	def get_parents(self, name):
		return [self._nodes[parent_name] for parent_name in self._parent_names[name]]

	@property
	def edge_style(self):
		"""
		:rtype: EdgeStyle
		"""
		return self._edge_style

	def __getitem__(self, item):
		if item in self._nodes:
			node = self._nodes[item]
			result = node.evaluate(self._echo)
			return result
		else:
			try:
				return super().__getitem__(item)
			except:
				raise AttributeError(item)



	def __getattr__(self, item):
		return self.__getitem__(item=item)

	def __getstate__(self):
		# fixes pickle and deepcopy
		return self.__dict__

	def __setstate__(self, state):
		# fixes pickle and deepcopy
		self.__dict__ = state

	@property
	def hard_memory(self):
		"""
		:rtype: HardMemory
		"""
		return self._hard_memory


	def add(
			self, name, function=None, obj=None, parents=None, evaluate=True, on_disk=False, node_style=None, edge_style=None,
			style_transferable=True
	):
		"""
		:type name: str
		:type parents: list[str]
		:type function: function
		:type evaluate: bool
		:type on_disk: bool
		:type node_style: NodeStyle or None
		"""
		if function is not None and obj is not None: raise ValueError('CG at least one of function and obj should be None!')
		if function is None: function = lambda: obj
		# Check inputs
		if not name:
			raise ValueError(f"Please specify a name for this node")
		if node_style is not None and type(node_style) is not NodeStyle:
			raise TypeError(f'node_style is {type(node_style)}')
		if edge_style is not None and type(edge_style) is not EdgeStyle:
			raise TypeError(f'edge_style is {type(edge_style)}')


		parents = parents or []
		parents = remove_list_duplicates(parents)

		if on_disk and self.hard_memory is None:
			self._hard_memory = HardMemory(path=self._hard_memory_path, log_access=True)

		# Check parent states are known
		unknown_states = set(parents).difference(set(self._nodes.keys()))
		if unknown_states:
			states_str = ', '.join([f"'{s}'" for s in unknown_states])
			raise KeyError(f"Error adding '{name}': Unknown parent state(s) {states_str}")

		# Create or update node

		parent_nodes = remove_list_duplicates([self._nodes[p] for p in parents])


		if name in self._nodes:
			node = self._nodes[name]
			node.update(
				parents=parent_nodes, function=function, on_disk=on_disk, style=node_style, edge_style=edge_style, echo=self._echo,
				style_transferable=style_transferable
			)
		else:
			if self._echo: print(f"CG Creating node '{name}'")
			node = ComputationGraphNode(
				name=name, parents=parent_nodes, function=function, graph=self, on_disk=on_disk, style=node_style, edge_style=edge_style,
				style_transferable=style_transferable, safe=self._safe
			)
			self._nodes[name] = node

		if evaluate:
			node = self._nodes[name]
			node.evaluate(self._echo)


	def delete(self, name):
		del self._nodes[name]



	def graphviz_str(self):

		dot_str = "strict digraph G { \n\t{\n " # {{node [margin={self.margin} fontcolor={self.font_colour} fontsize={self.font_size} width={self.width} shape={self.shape} style={self.style}] }}

		for n in self._nodes.values():
			node_str = f'{n.label} {n.style}'
			dot_str += f'\t\t{node_str}\n'
		dot_str += '\t}\n'
		for n in self._nodes.values():
			node_str = n.graphviz_edges_str
			if node_str:
				dot_str += f"\t{node_str}\n"
		dot_str += "}"

		return dot_str


	def __str__(self):
		if not len(self._nodes):
			return "<empty graph>"

		# Sort nodes topologically
		get_dependencies = lambda n: set(n.parent_names)
		toposort_in = { node.name: get_dependencies(node) for node in self._nodes.values() }
		toposorted = [l for g in toposort(toposort_in) for l in g]

		# Find longest strings so we can pad our strings to equal length later
		get_parents_str = lambda n: ', '.join(self._nodes[n].parent_names)
		data_to_print = [(get_parents_str(n), n, self._nodes[n].is_stale) for n in toposorted]
		longest_parents_str = max([len(d[0]) for d in data_to_print])
		longest_node_name = max([len(d[1]) for d in data_to_print])

		# Build output str in format: "[parents] --> node" along with an indicate of stale-ness
		result_str = ""
		for data in data_to_print:
			lside = data[0].rjust(longest_parents_str)
			rside = data[1].ljust(longest_node_name)
			stale_indic = " (stale)" if data[2] else ""
			result_str += f"{lside} --> {rside}{stale_indic}\n"
		result_str = result_str.rstrip('\n')
		return result_str

	def render(self):
		return graphviz.Source(self.graphviz_str())

class Snapshot:
	def __init__(self, graph):
		"""
		:type graph: ComputationGraph
		"""
		self._graphviz_str = graph.graphviz_str()

	def render(self):
		return graphviz.Source(self._graphviz_str)






from toposort import toposort
import graphviz
from .NodeStyle import NodeStyle
from .EdgeStyle import EdgeStyle
from .ComputationGraphNode import ComputationGraphNode
from memoria import HardMemory


class Graph:
	def __init__(self, echo=0, node_style=None, edge_style=None):
		"""
		:type echo: int
		:type node_style: NodeStyle or None
		"""
		echo = max(0, echo)
		self._echo = echo
		self._nodes = {}
		self._hard_memory = None
		original_node_style = NodeStyle(
			colour='gray80',
			text_colour='deepskyblue2',
			fill_colour='gray95',
			text_size=10,
			font='helvetica',
			shape='egg',
			style='filled'

		)
		if node_style is None:
			node_style = original_node_style
		else:
			node_style = node_style.complement(original_node_style)

		self._node_style = node_style

		original_edge_style = EdgeStyle(
			colour='grey60',
			arrow_size=0.5
		)
		if edge_style is None:
			edge_style = original_edge_style
		else:
			edge_style = edge_style.complement(original_edge_style)

		self._edge_style = edge_style

	@property
	def node_style(self):
		"""
		:rtype: NodeStyle
		"""
		return self._node_style

	@property
	def edge_style(self):
		"""
		:rtype: EdgeStyle
		"""
		return self._edge_style

	def __getitem__(self, item):
		node = self._nodes[item]
		result = node.evaluate(self._echo)
		return result

	def __getattr__(self, item):
		return self.__getitem__(item=item)

	@property
	def hard_memory(self):
		"""
		:rtype: HardMemory
		"""
		return self._hard_memory


	def add(self, name, parents, function, evaluate=True, on_disk=False, node_style=None):
		"""
		:type name: str
		:type parents: list[str]
		:type function: function
		:type evaluate: bool
		:type on_disk: bool
		:type node_style: NodeStyle or None
		"""

		# Check inputs
		if not name:
			raise ValueError(f"Please specify a name for this node")
		parents = parents or []
		if on_disk and self.hard_memory is None:
			self._hard_memory = HardMemory(path=self._hard_memory_path, log_access=True)

		# Check parent states are known
		unknown_states = set(parents).difference(set(self._nodes.keys()))
		if unknown_states:
			states_str = ', '.join([f"'{s}'" for s in unknown_states])
			raise KeyError(f"Error adding '{name}': Unknown parent state(s) {states_str}")

		# Create or update node
		parent_nodes = set([self._nodes[p] for p in parents])
		if name in self._nodes:
			node = self._nodes[name]
			node.update(parent_nodes, function, on_disk=on_disk, style=node_style, echo=self._echo)
		else:
			if self._echo: print(f"CG Creating node '{name}'")
			node = ComputationGraphNode(name, parent_nodes, function, graph=self, on_disk=on_disk, style=node_style)
			self._nodes[name] = node

		if evaluate:
			node = self._nodes[name]
			node.evaluate(self._echo)


	def delete(self, name):
		del self._nodes[name]



	def graphviz_str(self):
		def node_to_graphviz(n):
			"""
			:type n: ComputationGraphNode
			:rtype: str
			"""
			#node_name = f"\"{n.name}\""
			node_name_and_status = f"\"{n.name_and_status}\""
			parent_names = [f"\"{p.name_and_status}\"" for p in n.parent_names]
			if not len(parent_names):
				return node_name_and_status if not len(n.child_names) else None
			if len(parent_names) == 1:
				return f"{parent_names[0]} -> {node_name_and_status} {str(self.edge_style)}"
			else:
				return f"{{ {' '.join(parent_names)} }} -> {node_name_and_status} {str(self.edge_style)}"

		dot_str = "strict digraph G { \n\t{\n " # {{node [margin={self.margin} fontcolor={self.font_colour} fontsize={self.font_size} width={self.width} shape={self.shape} style={self.style}] }}

		for n in self._nodes.values():
			node_str = f'"{n.name_and_status}" {n.style}'
			dot_str += f'\t\t{node_str}\n'
		dot_str += '\t}\n'
		for n in self._nodes.values():
			node_str = node_to_graphviz(n)
			if node_str:
				dot_str += f"\t{node_str}\n"
		dot_str += "}"

		return dot_str


	def __str__(self):
		if not len(self._nodes):
			return "<empty graph>"

		# Sort nodes topologically
		get_dependencies = lambda n: set([c.name for c in n.parent_names])
		toposort_in = { node.name: get_dependencies(node) for node in self._nodes.values() }
		toposorted = [l for g in toposort(toposort_in) for l in g]

		# Find longest strings so we can pad our strings to equal length later
		get_parents_str = lambda n: ', '.join([p.name for p in self._nodes[n].parent_names])
		data_to_print = [(get_parents_str(n), n, self._nodes[n].is_stale) for n in toposorted]
		longest_parents_str = max([len(d[0]) for d in data_to_print])
		longest_node_name = max([len(d[1]) for d in data_to_print])

		# Build output str in format: "[parents] --> node" along with an indicate of stale-ness
		result_str = ""
		for data in data_to_print:
			lside = data[0].rjust(longest_parents_str)
			rside = data[1].ljust(longest_node_name)
			stale_indic = " (stale)" if data[2] else ""
			result_str += f"{lside} --> {rside}{stale_indic}\n"
		result_str = result_str.rstrip('\n')
		return result_str

	def render(self):
		return graphviz.Source(self.graphviz_str())






from .NodeStyle import NodeStyle
from .EdgeStyle import EdgeStyle
from slytherin.collections import remove_list_duplicates
from copy import deepcopy
from memoria import Pickler, make_hash_sha256
import dill

class ComputationGraphNode():
	def __init__(
			self, name, function, graph, parents=None, on_disk=False, style=None, edge_style=None, style_transferable=True,
			safe=True,
	):
		"""
		:type name: str
		:type parents: list[ComputationGraphNode]
		:type graph: ComputationGraph
		:type function: function
		:type graph: ComputationGraph
		:type on_disk: bool
		:type style: NodeStyle or None
		:type input_edge_style: EdgeStyle
		:type output_edge_style: EdgeStyle
		:type safe: bool
		:param safe: when True only a copy of the value is returned to avoid mutating it from outside
		"""
		# make parents unique
		parents = parents or []
		self._name = name
		#self._parents = []
		#self._children = []
		self._graph = graph
		self._value = None
		self._style = None
		self._input_edge_style = None
		self._output_edge_style = None
		self._style_transferable = None
		self._safe = safe
		if self.name not in self.graph._child_names:
			self.graph._child_names[self.name] = []
		if self.name not in self.graph._parent_names:
			self.graph._parent_names[self.name] = []
		self._frozen = False
		self._stale = True
		self._function = None
		self._on_disk = on_disk

		self.update(
			parents, function, on_disk=on_disk, style=style, edge_style=edge_style,
			style_transferable=style_transferable
		)

	def __eq__(self, other):
		return isinstance(other, ComputationGraphNode) and self.name == other.name

	def __hash__(self):
		return hash(self.name)

	def __repr__(self):
		return f'ComputationGraphNode:{self.name}'

	def __getstate__(self):
		# fixes pickle and deepcopy
		new_dict = dict(self.__dict__)
		new_dict['_function'] = dill.dumps(obj=new_dict['_function'])
		return new_dict

	def __setstate__(self, state):
		# fixes pickle and deepcopy
		self.__dict__ = state
		try:
			self.__dict__['_function'] = dill.loads(str=self.__dict__['_function'])
		except:
			pass

	def save(self, path, echo=1):
		"""
		:type path: str
		"""
		echo = max(0, echo)
		graph = self._graph
		value = self._value
		stale = self._stale
		frozen = self._frozen
		self._graph = None
		if echo: print(f'GN saving node "{self.name}" ', end='')
		try:
			Pickler.save(path=path, obj=self, echo=echo-1)
			if echo: print('done!')
		except Exception as e:
			# If the value is hard to save, we save the node without the value
			# but we make the node stale so that
			# next time it is evaluated it recreates its value
			if echo: print(e, ' ', end='')
			self._value = None
			self._stale = True
			self._frozen = False
			Pickler.save(path=path, obj=self, echo=echo-1)
			if echo: print('done!')
		finally:
			self._graph = graph
			self._value = value
			self._stale = stale
			self._frozen = frozen


	@staticmethod
	def load(path, graph, echo=0):
		"""
		:type path: str
		:type graph: ComputationGraph
		:rtype: ComputationGraphNode
		"""
		echo = max(0, echo)
		node = Pickler.load(path=path, echo=echo)
		node._graph = graph
		return node

	@property
	def is_frozen(self):
		return self._frozen

	@property
	def is_stale(self):
		return self._stale

	def freeze(self):
		self._frozen = True

	def unfreeze(self):
		self._frozen = False
		if self._stale:
			self.mark_stale()

	@property
	def graph(self):
		"""
		:rtype: ComputationGraph
		"""
		return self._graph


	@property
	def name(self):
		return self._name

	@property
	def label(self):
		name = self.name.replace('__', '\n').replace('_', ' ')
		if self.is_stale and self.is_frozen:
			return f'"{name}\n(stale & frozen)"'
		elif self.is_stale and not self.is_frozen:
			return f'"{name}\n(stale)"'
		elif not self.is_stale and self.is_frozen:
			return f'"{name}\n(frozen)"'
		else:
			return f'"{name}"'

	@property
	def parent_names(self):
		"""
		:type: list[str]
		"""
		return list(self._graph._parent_names[self.name])

	@property
	def parents(self):
		return self.graph.get_parents(name=self.name)

	@property
	def has_parents(self):
		return len(self.parent_names) > 0

	@property
	def child_names(self):
		"""
		:type: list[str]
		"""
		return list(self._graph._child_names[self.name])

	@property
	def children(self):
		return self.graph.get_children(name=self.name)

	@property
	def has_children(self):
		return len(self.child_names) > 0

	def remove_child(self, child):
		if isinstance(child, str):
			self.graph._child_names[self.name].remove(child)
		else:
			self.graph._child_names[self.name].remove(child.name)

	############################################################## COMPUTATION

	def update(self, parents, function, on_disk, style=None, edge_style=None, echo=0, style_transferable=True):
		"""
		:type parents: list[ComputationGraphNode]
		:type function: function
		:type on_disk: bool
		:type style: NodeStyle
		:type echo:
		:return:
		"""
		# make parents unique:
		parents = parents or []
		parents = remove_list_duplicates(parents)
		echo = max(0, echo)
		if echo: print(f"GN Updating '{self.name}'")
		if self._on_disk and not on_disk:
			self._graph.hard_memory.delete(key=self.name)
		self._on_disk = on_disk

		if style is not None:
			if style is False:
				self._style = None
			else:
				self._style = style

		if edge_style is not None:
			if edge_style is False:
				self._input_edge_style = None
				self._output_edge_style = None
			else:
				self._input_edge_style = edge_style
				self._output_edge_style = edge_style
		self._style_transferable = style_transferable

		parent_names = [p.name for p in parents]

		removed_parent_names = [name for name in self.parent_names if name not in parent_names] # self.unique_parents.difference(parents)
		new_parent_names = [name for name in parent_names if name not in self.parent_names]# set(parents).difference(self._parents)
		#self._parents = parents

		self.graph._parent_names[self.name] = parent_names
		for parent_name in removed_parent_names:
			self.graph._child_names[parent_name].remove(self.name)
		for parent_name in new_parent_names:
			self.graph._child_names[parent_name].append(self.name)

		self._function = function
		self.mark_stale()

	@property
	def value(self):
		return self._value

	@value.setter
	def value(self, value):
		if self.is_frozen:
			raise RuntimeError('GN: You cannot change the value while the node is frozen!')

		# only mark children stale if the value changes
		if make_hash_sha256(value) != make_hash_sha256(self._value):
			for child in self.children:
				child.mark_stale()

		if self._on_disk:
			self._graph.hard_memory.save(obj=value, key=self.name, echo=1, text=self.name)
		else:
			self._value = value

	def mark_stale(self):
		self._stale = True

	def evaluate(self, echo=1):
		echo = max(0, echo)
		if not self.is_stale or self.is_frozen:
			return self.value

		parent_vals = {p.name: p.evaluate(echo) for p in self.parents}
		if echo: print(f"GN Evaluating '{self.name}'")

		self._stale = False

		if len(self.parent_names) == 0:
			self.value = self._function()
		elif len(self.parent_names) == 1:
			parent_val = list(parent_vals.values())[0]
			self.value = self._function(parent_val)
		else:
			inputs = GraphEvaluationInput(parent_vals)
			self.value = self._function(inputs)

		self._hash = make_hash_sha256(self.value)
		return self.value




	@property
	def graphviz_edges_str(self):
		if not self.has_parents:
			return self.label if not self.has_children else None
		else:
			edges = [
				f'{parent.label} -> {self.label} {str(self.get_edge_to_parent_style(parent=parent))}'
				for parent in self.parents
			]
			return '\n'.join(edges)

	############################################################## STYLE

	@property
	def parents_with_transferable_style(self):
		result = []
		for parent in self.parents:
			if parent._style_transferable:
				result.append(parent)
			elif len(parent.parents_with_transferable_style)>0:
				result.append(parent.parents_with_transferable_style[0])
		return result


	@property
	def inherited_style(self):
		"""
		:rtype: NodeStyle
		"""
		if self.has_parents:
			parents_with_style = [
				parent for parent in self.parents_with_transferable_style if parent.inherited_style is not None
			]
			if len(parents_with_style) > 0:
				parent_style = parents_with_style[0].inherited_style
			else:
				parent_style = None
		else:
			parent_style = None



		if self._style is None:
			return parent_style
		elif type(self._style) is not NodeStyle:
			raise TypeError(f'node.style of {self.name} is not NodeStyle! it is {type(self._style)}')
		else:
			return self._style.complement(parent_style)
	@property
	def style(self):
		if self.inherited_style is None:
			return self.graph.node_style
		else:
			return self.inherited_style.complement(self.graph.node_style)

	@property
	def inherited_output_edge_style(self):
		"""
		:rtype: EdgeStyle
		"""
		if self.has_parents:
			parents_with_edge_style = [parent for parent in self.parents if parent.inherited_output_edge_style is not None]
			if len(parents_with_edge_style) > 0:
				parent_output_edge_style = parents_with_edge_style[0].inherited_output_edge_style
			else:
				parent_output_edge_style = None
		else:
			parent_output_edge_style = None

		if self._output_edge_style is None:
			return parent_output_edge_style
		elif type(self._output_edge_style) is not EdgeStyle:
			raise TypeError('node._output_edge_style is not EdgeStyle')
		else:
			return self._output_edge_style.complement(parent_output_edge_style)

	def get_edge_to_parent_style(self, parent):
		if parent.inherited_output_edge_style is not None:
			return parent.inherited_output_edge_style.complement(self.graph.edge_style)

		else:
			return self.graph.edge_style

	@property
	def parent_edge_style(self):
		raise ValueError('should not use this')

	@parent_edge_style.setter
	def parent_edge_style(self, parent_edge_style):
		self._input_edge_style = parent_edge_style

	@property
	def child_edge_style(self):
		raise ValueError('should not be used')

	@child_edge_style.setter
	def child_edge_style(self, child_edge_style):
		self._output_edge_style = child_edge_style




class GraphEvaluationInput():
	def __init__(self, inputs):
		self.__dict__ = inputs

	def __getitem__(self, name):
		return self.__dict__[name]


from .ComputationGraph import Snapshot
from .ComputationGraph import ComputationGraph as Pensieve
from .ComputationGraph import ComputationGraph
from .NodeStyle import NodeStyle
from .EdgeStyle import EdgeStyle


from setuptools import setup, find_packages

def readme():
    with open('./README.md') as f:
        return f.read()

setup(
    name='pensieve',
    version='2.0.1',
    license='GNU AGPLv3',

    author='Idin',
    author_email='py@idin.ca',
    url='https://github.com/idin/pensieve',

    keywords='graph computation',
    description='Implementation of a computation graph',
    long_description=readme(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],

    packages=find_packages(exclude=("jupyter_tests", ".idea", ".git")),
    install_requires=['toposort', 'graphviz', 'memoria'],
    python_requires='~=3.6',
    zip_safe=True,
    test_suite='nose.collector',
    tests_require=['nose', 'coverage']
)
